@@ -27,7 +27,6 @@ from typing import (
 import libcst as cst
 from libcst import ensure_type
 from libcst._add_slots import add_slots
-from libcst._metadata_dependent import MetadataDependent
 from libcst.helpers import get_full_name_for_node
 from libcst.metadata.base_provider import BatchableMetadataProvider
 from libcst.metadata.expression_context_provider import (
@@ -35,7 +34,6 @@ from libcst.metadata.expression_context_provider import (
     ExpressionContextProvider,
 )
 
-
 # Comprehensions are handled separately in _visit_comp_alike due to
 # the complexity of the semantics
 _ASSIGNMENT_LIKE_NODES = (
@@ -45,7 +43,6 @@ _ASSIGNMENT_LIKE_NODES = (
     cst.AugAssign,
     cst.ClassDef,
     cst.CompFor,
-    cst.For,
     cst.FunctionDef,
     cst.Global,
     cst.Import,
@@ -76,9 +73,10 @@ class Access:
 
     #: The node of the access. A name is an access when the expression context is
     #: :attr:`ExpressionContext.LOAD`. This is usually the name node representing the
-    #: access, except for dotted imports, when it might be the attribute that
-    #: represents the most specific part of the imported symbol.
-    node: Union[cst.Name, cst.Attribute]
+    #: access, except for: 1) dotted imports, when it might be the attribute that
+    #: represents the most specific part of the imported symbol; and 2) string
+    #: annotations, when it is the entire string literal
+    node: Union[cst.Name, cst.Attribute, cst.BaseString]
 
     #: The scope of the access. Note that a access could be in a child scope of its
     #: assignment.
@@ -125,11 +123,28 @@ class Access:
             for assignment in assignments
             if assignment.scope != self.scope or assignment._index < self.__index
         }
-        if not previous_assignments and assignments:
+        if not previous_assignments and assignments and self.scope.parent != self.scope:
             previous_assignments = self.scope.parent[name]
         self.__assignments |= previous_assignments
 
 
+class QualifiedNameSource(Enum):
+    IMPORT = auto()
+    BUILTIN = auto()
+    LOCAL = auto()
+
+
+@add_slots
+@dataclass(frozen=True)
+class QualifiedName:
+    #: Qualified name, e.g. ``a.b.c`` or ``fn.<locals>.var``.
+    name: str
+
+    #: Source of the name, either :attr:`QualifiedNameSource.IMPORT`, :attr:`QualifiedNameSource.BUILTIN`
+    #: or :attr:`QualifiedNameSource.LOCAL`.
+    source: QualifiedNameSource
+
+
 class BaseAssignment(abc.ABC):
     """Abstract base class of :class:`Assignment` and :class:`BuitinAssignment`."""
 
@@ -177,6 +192,10 @@ class BaseAssignment(abc.ABC):
         """Return an integer that represents the order of assignments in `scope`"""
         return -1
 
+    @abc.abstractmethod
+    def get_qualified_names_for(self, full_name: str) -> Set[QualifiedName]:
+        ...
+
 
 class Assignment(BaseAssignment):
     """An assignment records the name, CSTNode and its accesses."""
@@ -197,6 +216,26 @@ class Assignment(BaseAssignment):
     def _index(self) -> int:
         return self.__index
 
+    def get_qualified_names_for(self, full_name: str) -> Set[QualifiedName]:
+        scope = self.scope
+        name_prefixes = []
+        while scope:
+            if isinstance(scope, ClassScope):
+                name_prefixes.append(scope.name)
+            elif isinstance(scope, FunctionScope):
+                name_prefixes.append(f"{scope.name}.<locals>")
+            elif isinstance(scope, ComprehensionScope):
+                name_prefixes.append("<comprehension>")
+            elif not isinstance(scope, (GlobalScope, BuiltinScope)):
+                raise Exception(f"Unexpected Scope: {scope}")
+
+            scope = scope.parent if scope.parent != scope else None
+
+        parts = [*reversed(name_prefixes)]
+        if full_name:
+            parts.append(full_name)
+        return {QualifiedName(".".join(parts), QualifiedNameSource.LOCAL)}
+
 
 # even though we don't override the constructor.
 class BuiltinAssignment(BaseAssignment):
@@ -207,7 +246,74 @@ class BuiltinAssignment(BaseAssignment):
     `types <https://docs.python.org/3/library/stdtypes.html>`_.
     """
 
-    pass
+    def get_qualified_names_for(self, full_name: str) -> Set[QualifiedName]:
+        return {QualifiedName(f"builtins.{self.name}", QualifiedNameSource.BUILTIN)}
+
+
+class ImportAssignment(Assignment):
+    """An assignment records the import node and it's alias"""
+
+    as_name: cst.CSTNode
+
+    def __init__(
+        self,
+        name: str,
+        scope: "Scope",
+        node: cst.CSTNode,
+        index: int,
+        as_name: cst.CSTNode,
+    ):
+        super().__init__(name, scope, node, index)
+        self.as_name = as_name
+
+    def get_module_name_for_import(self) -> str:
+        module = ""
+        if isinstance(self.node, cst.ImportFrom):
+            module_attr = self.node.module
+            relative = self.node.relative
+            if module_attr:
+                module = get_full_name_for_node(module_attr) or ""
+            if relative:
+                module = "." * len(relative) + module
+        return module
+
+    def get_qualified_names_for(self, full_name: str) -> Set[QualifiedName]:
+        module = self.get_module_name_for_import()
+        results = set()
+        import_names = self.node.names
+        if not isinstance(import_names, cst.ImportStar):
+            for name in import_names:
+                real_name = get_full_name_for_node(name.name)
+                if not real_name:
+                    continue
+                # real_name can contain `.` for dotted imports
+                # for these we want to find the longest prefix that matches full_name
+                parts = real_name.split(".")
+                real_names = [".".join(parts[:i]) for i in range(len(parts), 0, -1)]
+                for real_name in real_names:
+                    as_name = real_name
+                    if module and module.endswith("."):
+                        # from . import a
+                        # real_name should be ".a"
+                        real_name = f"{module}{real_name}"
+                    elif module:
+                        real_name = f"{module}.{real_name}"
+                    if name and name.asname:
+                        eval_alias = name.evaluated_alias
+                        if eval_alias is not None:
+                            as_name = eval_alias
+                    if full_name.startswith(as_name):
+                        remaining_name = full_name.split(as_name, 1)[1].lstrip(".")
+                        results.add(
+                            QualifiedName(
+                                f"{real_name}.{remaining_name}"
+                                if remaining_name
+                                else real_name,
+                                QualifiedNameSource.IMPORT,
+                            )
+                        )
+                        break
+        return results
 
 
 class Assignments:
@@ -254,23 +360,6 @@ class Accesses:
         return len(self[node]) > 0
 
 
-class QualifiedNameSource(Enum):
-    IMPORT = auto()
-    BUILTIN = auto()
-    LOCAL = auto()
-
-
-@add_slots
-@dataclass(frozen=True)
-class QualifiedName:
-    #: Qualified name, e.g. ``a.b.c`` or ``fn.<locals>.var``.
-    name: str
-
-    #: Source of the name, either :attr:`QualifiedNameSource.IMPORT`, :attr:`QualifiedNameSource.BUILTIN`
-    #: or :attr:`QualifiedNameSource.LOCAL`.
-    source: QualifiedNameSource
-
-
 class _NameUtil:
     @staticmethod
     def get_name_for(node: Union[str, cst.CSTNode]) -> Optional[str]:
@@ -287,72 +376,6 @@ class _NameUtil:
             return _NameUtil.get_name_for(node.name)
         return None
 
-    @staticmethod
-    def find_qualified_name_for_import_alike(
-        assignment_node: Union[cst.Import, cst.ImportFrom], full_name: str
-    ) -> Set[QualifiedName]:
-        module = ""
-        results = set()
-        if isinstance(assignment_node, cst.ImportFrom):
-            module_attr = assignment_node.module
-            if module_attr:
-                # TODO: for relative import, keep the relative Dot in the qualified name
-                module = get_full_name_for_node(module_attr)
-        import_names = assignment_node.names
-        if not isinstance(import_names, cst.ImportStar):
-            for name in import_names:
-                real_name = get_full_name_for_node(name.name)
-                if not real_name:
-                    continue
-                # real_name can contain `.` for dotted imports
-                # for these we want to find the longest prefix that matches full_name
-                parts = real_name.split(".")
-                real_names = [".".join(parts[:i]) for i in range(len(parts), 0, -1)]
-                for real_name in real_names:
-                    as_name = real_name
-                    if module:
-                        real_name = f"{module}.{real_name}"
-                    if name and name.asname:
-                        eval_alias = name.evaluated_alias
-                        if eval_alias is not None:
-                            as_name = eval_alias
-                    if full_name.startswith(as_name):
-                        remaining_name = full_name.split(as_name, 1)[1].lstrip(".")
-                        results.add(
-                            QualifiedName(
-                                f"{real_name}.{remaining_name}"
-                                if remaining_name
-                                else real_name,
-                                QualifiedNameSource.IMPORT,
-                            )
-                        )
-                        break
-        return results
-
-    @staticmethod
-    def find_qualified_name_for_non_import(
-        assignment: Assignment, remaining_name: str
-    ) -> Set[QualifiedName]:
-        scope = assignment.scope
-        name_prefixes = []
-        while scope:
-            if isinstance(scope, ClassScope):
-                name_prefixes.append(scope.name)
-            elif isinstance(scope, FunctionScope):
-                name_prefixes.append(f"{scope.name}.<locals>")
-            elif isinstance(scope, GlobalScope):
-                break
-            elif isinstance(scope, ComprehensionScope):
-                name_prefixes.append("<comprehension>")
-            else:
-                raise Exception(f"Unexpected Scope: {scope}")
-            scope = scope.parent
-
-        parts = [*reversed(name_prefixes)]
-        if remaining_name:
-            parts.append(remaining_name)
-        return {QualifiedName(".".join(parts), QualifiedNameSource.LOCAL)}
-
 
 class Scope(abc.ABC):
     """
@@ -391,10 +414,33 @@ class Scope(abc.ABC):
         self._assignment_count = 0
 
     def record_assignment(self, name: str, node: cst.CSTNode) -> None:
-        self._assignments[name].add(
-            Assignment(name=name, scope=self, node=node, index=self._assignment_count)
+        target = self._find_assignment_target(name)
+        target._assignments[name].add(
+            Assignment(
+                name=name, scope=target, node=node, index=target._assignment_count
+            )
         )
 
+    def record_import_assignment(
+        self, name: str, node: cst.CSTNode, as_name: cst.CSTNode
+    ) -> None:
+        target = self._find_assignment_target(name)
+        target._assignments[name].add(
+            ImportAssignment(
+                name=name,
+                scope=target,
+                node=node,
+                as_name=as_name,
+                index=target._assignment_count,
+            )
+        )
+
+    def _find_assignment_target(self, name: str) -> "Scope":
+        return self
+
+    def _find_assignment_target_parent(self, name: str) -> "Scope":
+        return self
+
     def record_access(self, name: str, access: Access) -> None:
         self._accesses[name].add(access)
 
@@ -406,13 +452,9 @@ class Scope(abc.ABC):
         """Overridden by ClassScope to hide it's assignments from child scopes."""
         return name in self
 
-    def _record_assignment_as_parent(self, name: str, node: cst.CSTNode) -> None:
-        """Overridden by ClassScope to forward 'nonlocal' assignments from child scopes."""
-        self.record_assignment(name, node)
-
     @abc.abstractmethod
     def __contains__(self, name: str) -> bool:
-        """ Check if the name str exist in current scope by ``name in scope``. """
+        """Check if the name str exist in current scope by ``name in scope``."""
         ...
 
     @abc.abstractmethod
@@ -509,22 +551,14 @@ class Scope(abc.ABC):
                 assignments = self[prefix]
                 break
         for assignment in assignments:
-            if isinstance(assignment, Assignment):
-                assignment_node = assignment.node
-                if isinstance(assignment_node, (cst.Import, cst.ImportFrom)):
-                    results |= _NameUtil.find_qualified_name_for_import_alike(
-                        assignment_node, full_name
-                    )
-                else:
-                    results |= _NameUtil.find_qualified_name_for_non_import(
-                        assignment, full_name
-                    )
-            elif isinstance(assignment, BuiltinAssignment):
-                results.add(
-                    QualifiedName(
-                        f"builtins.{assignment.name}", QualifiedNameSource.BUILTIN
-                    )
-                )
+            names = assignment.get_qualified_names_for(full_name)
+            if (
+                isinstance(assignment, Assignment)
+                and not isinstance(node, str)
+                and _is_assignment(node, assignment.node)
+            ):
+                return names
+            results |= names
         return results
 
     @property
@@ -538,27 +572,57 @@ class Scope(abc.ABC):
         return Accesses(self._accesses)
 
 
+class BuiltinScope(Scope):
+    """
+    A BuiltinScope represents python builtin declarations. See https://docs.python.org/3/library/builtins.html
+    """
+
+    def __init__(self, globals: Scope) -> None:
+        self.globals: Scope = globals  # must be defined before Scope.__init__ is called
+        super().__init__(parent=self)
+
+    def __contains__(self, name: str) -> bool:
+        return hasattr(builtins, name)
+
+    def __getitem__(self, name: str) -> Set[BaseAssignment]:
+        if name in self._assignments:
+            return self._assignments[name]
+        if hasattr(builtins, name):
+            # note - we only see the builtin assignments during the deferred
+            # access resolution. unfortunately that means we have to create the
+            # assignment here, which can cause the set to mutate during iteration
+            self._assignments[name].add(BuiltinAssignment(name, self))
+            return self._assignments[name]
+        return set()
+
+    def record_global_overwrite(self, name: str) -> None:
+        raise NotImplementedError("global overwrite in builtin scope are not allowed")
+
+    def record_nonlocal_overwrite(self, name: str) -> None:
+        raise NotImplementedError("declarations in builtin scope are not allowed")
+
+    def _find_assignment_target(self, name: str) -> "Scope":
+        raise NotImplementedError("assignments in builtin scope are not allowed")
+
+
 class GlobalScope(Scope):
     """
     A GlobalScope is the scope of module. All module level assignments are recorded in GlobalScope.
     """
 
     def __init__(self) -> None:
-        self.globals: Scope = self  # must be defined before Scope.__init__ is called
-        super().__init__(parent=self)
+        super().__init__(parent=BuiltinScope(self))
 
     def __contains__(self, name: str) -> bool:
-        return hasattr(builtins, name) or (
-            name in self._assignments and len(self._assignments[name]) > 0
-        )
+        if name in self._assignments:
+            return len(self._assignments[name]) > 0
+        return self.parent._contains_in_self_or_parent(name)
 
     def __getitem__(self, name: str) -> Set[BaseAssignment]:
-        if hasattr(builtins, name):
-            if not any(
-                isinstance(i, BuiltinAssignment) for i in self._assignments[name]
-            ):
-                self._assignments[name].add(BuiltinAssignment(name, self))
-        return self._assignments[name]
+        if name in self._assignments:
+            return self._assignments[name]
+        else:
+            return self.parent._getitem_from_self_or_parent(name)
 
     def record_global_overwrite(self, name: str) -> None:
         pass
@@ -590,11 +654,11 @@ class LocalScope(Scope, abc.ABC):
     def record_nonlocal_overwrite(self, name: str) -> None:
         self._scope_overwrites[name] = self.parent
 
-    def record_assignment(self, name: str, node: cst.CSTNode) -> None:
+    def _find_assignment_target(self, name: str) -> "Scope":
         if name in self._scope_overwrites:
-            self._scope_overwrites[name]._record_assignment_as_parent(name, node)
+            return self._scope_overwrites[name]._find_assignment_target_parent(name)
         else:
-            super().record_assignment(name, node)
+            return super()._find_assignment_target(name)
 
     def __contains__(self, name: str) -> bool:
         if name in self._scope_overwrites:
@@ -627,7 +691,7 @@ class ClassScope(LocalScope):
     When a class is defined, it creates a ClassScope.
     """
 
-    def _record_assignment_as_parent(self, name: str, node: cst.CSTNode) -> None:
+    def _find_assignment_target_parent(self, name: str) -> "Scope":
         """
         Forward the assignment to parent.
 
@@ -642,7 +706,7 @@ class ClassScope(LocalScope):
                                  # hidden from its children.
 
         """
-        self.parent._record_assignment_as_parent(name, node)
+        return self.parent._find_assignment_target_parent(name)
 
     def _getitem_from_self_or_parent(self, name: str) -> Set[BaseAssignment]:
         """
@@ -707,18 +771,48 @@ def _gen_dotted_names(
                 yield from name_values
 
 
+def _is_assignment(node: cst.CSTNode, assignment_node: cst.CSTNode) -> bool:
+    """
+    Returns true if ``node`` is part of the assignment at ``assignment_node``.
+
+    Normally this is just a simple identity check, except for imports where the
+    assignment is attached to the entire import statement but we are interested in
+    ``Name`` nodes inside the statement.
+    """
+    if node is assignment_node:
+        return True
+    if isinstance(assignment_node, (cst.Import, cst.ImportFrom)):
+        aliases = assignment_node.names
+        if isinstance(aliases, cst.ImportStar):
+            return False
+        for alias in aliases:
+            if alias.name is node:
+                return True
+            asname = alias.asname
+            if asname is not None:
+                if asname.name is node:
+                    return True
+    return False
+
+
+@dataclass(frozen=True)
+class DeferredAccess:
+    access: Access
+    enclosing_attribute: Optional[cst.Attribute]
+    enclosing_string_annotation: Optional[cst.BaseString]
+
+
 class ScopeVisitor(cst.CSTVisitor):
     # since it's probably not useful. That can makes this visitor cleaner.
     def __init__(self, provider: "ScopeProvider") -> None:
         self.provider: ScopeProvider = provider
         self.scope: Scope = GlobalScope()
-        self.__deferred_accesses: List[Tuple[Access, Optional[cst.Attribute]]] = []
+        self.__deferred_accesses: List[DeferredAccess] = []
         self.__top_level_attribute_stack: List[Optional[cst.Attribute]] = [None]
-        self.__in_annotation: Set[
-            Union[cst.Call, cst.Annotation, cst.Subscript]
-        ] = set()
-        self.__in_type_hint: Set[Union[cst.Call, cst.Annotation, cst.Subscript]] = set()
+        self.__in_annotation_stack: List[bool] = [False]
+        self.__in_type_hint_stack: List[bool] = [False]
         self.__in_ignored_subscript: Set[cst.Subscript] = set()
+        self.__last_string_annotation: Optional[cst.BaseString] = None
         self.__ignore_annotation: int = 0
 
     @contextmanager
@@ -752,11 +846,15 @@ class ScopeVisitor(cst.CSTVisitor):
             asname = name.asname
             if asname is not None:
                 name_values = _gen_dotted_names(cst.ensure_type(asname.name, cst.Name))
+                import_node_asname = asname.name
             else:
                 name_values = _gen_dotted_names(name.name)
+                import_node_asname = name.name
 
             for name_value, _ in name_values:
-                self.scope.record_assignment(name_value, node)
+                self.scope.record_import_assignment(
+                    name_value, node, import_node_asname
+                )
         return False
 
     def visit_Import(self, node: cst.Import) -> Optional[bool]:
@@ -775,19 +873,21 @@ class ScopeVisitor(cst.CSTVisitor):
 
     def visit_Call(self, node: cst.Call) -> Optional[bool]:
         self.__top_level_attribute_stack.append(None)
+        self.__in_type_hint_stack.append(False)
+        self.__in_annotation_stack.append(False)
         qnames = {qn.name for qn in self.scope.get_qualified_names_for(node)}
         if "typing.NewType" in qnames or "typing.TypeVar" in qnames:
             node.func.visit(self)
-            self.__in_type_hint.add(node)
+            self.__in_type_hint_stack[-1] = True
             for arg in node.args[1:]:
                 arg.visit(self)
             return False
         if "typing.cast" in qnames:
             node.func.visit(self)
             if len(node.args) > 0:
-                self.__in_type_hint.add(node)
+                self.__in_type_hint_stack.append(True)
                 node.args[0].visit(self)
-                self.__in_type_hint.discard(node)
+                self.__in_type_hint_stack.pop()
                 for arg in node.args[1:]:
                     arg.visit(self)
             return False
@@ -795,13 +895,14 @@ class ScopeVisitor(cst.CSTVisitor):
 
     def leave_Call(self, original_node: cst.Call) -> None:
         self.__top_level_attribute_stack.pop()
-        self.__in_type_hint.discard(original_node)
+        self.__in_type_hint_stack.pop()
+        self.__in_annotation_stack.pop()
 
     def visit_Annotation(self, node: cst.Annotation) -> Optional[bool]:
-        self.__in_annotation.add(node)
+        self.__in_annotation_stack.append(True)
 
     def leave_Annotation(self, original_node: cst.Annotation) -> None:
-        self.__in_annotation.discard(original_node)
+        self.__in_annotation_stack.pop()
 
     def visit_SimpleString(self, node: cst.SimpleString) -> Optional[bool]:
         self._handle_string_annotation(node)
@@ -815,25 +916,39 @@ class ScopeVisitor(cst.CSTVisitor):
     ) -> bool:
         """Returns whether it successfully handled the string annotation"""
         if (
-            self.__in_type_hint or self.__in_annotation
+            self.__in_type_hint_stack[-1] or self.__in_annotation_stack[-1]
         ) and not self.__in_ignored_subscript:
             value = node.evaluated_value
             if value:
-                mod = cst.parse_module(value)
-                mod.visit(self)
+                top_level_annotation = self.__last_string_annotation is None
+                if top_level_annotation:
+                    self.__last_string_annotation = node
+                try:
+                    mod = cst.parse_module(value)
+                    mod.visit(self)
+                except cst.ParserSyntaxError:
+                    # swallow string annotation parsing errors
+                    # this is the same behavior as cPython
+                    pass
+                if top_level_annotation:
+                    self.__last_string_annotation = None
                 return True
         return False
 
     def visit_Subscript(self, node: cst.Subscript) -> Optional[bool]:
-        qnames = {qn.name for qn in self.scope.get_qualified_names_for(node.value)}
-        if any(qn.startswith(("typing.", "typing_extensions.")) for qn in qnames):
-            self.__in_type_hint.add(node)
-        if "typing.Literal" in qnames or "typing_extensions.Literal" in qnames:
-            self.__in_ignored_subscript.add(node)
+        in_type_hint = False
+        if isinstance(node.value, cst.Name):
+            qnames = {qn.name for qn in self.scope.get_qualified_names_for(node.value)}
+            if any(qn.startswith(("typing.", "typing_extensions.")) for qn in qnames):
+                in_type_hint = True
+            if "typing.Literal" in qnames or "typing_extensions.Literal" in qnames:
+                self.__in_ignored_subscript.add(node)
+
+        self.__in_type_hint_stack.append(in_type_hint)
         return True
 
     def leave_Subscript(self, original_node: cst.Subscript) -> None:
-        self.__in_type_hint.discard(original_node)
+        self.__in_type_hint_stack.pop()
         self.__in_ignored_subscript.discard(original_node)
 
     def visit_Name(self, node: cst.Name) -> Optional[bool]:
@@ -846,12 +961,16 @@ class ScopeVisitor(cst.CSTVisitor):
                 node,
                 self.scope,
                 is_annotation=bool(
-                    self.__in_annotation and not self.__ignore_annotation
+                    self.__in_annotation_stack[-1] and not self.__ignore_annotation
                 ),
-                is_type_hint=bool(self.__in_type_hint),
+                is_type_hint=bool(self.__in_type_hint_stack[-1]),
             )
             self.__deferred_accesses.append(
-                (access, self.__top_level_attribute_stack[-1])
+                DeferredAccess(
+                    access=access,
+                    enclosing_attribute=self.__top_level_attribute_stack[-1],
+                    enclosing_string_annotation=self.__last_string_annotation,
+                )
             )
 
     def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
@@ -992,12 +1111,25 @@ class ScopeVisitor(cst.CSTVisitor):
                 node.elt.visit(self)
         return False
 
+    def visit_For(self, node: cst.For) -> Optional[bool]:
+        node.target.visit(self)
+        self.scope._assignment_count += 1
+        for child in [node.iter, node.body, node.orelse, node.asynchronous]:
+            if child is not None:
+                child.visit(self)
+        return False
+
     def infer_accesses(self) -> None:
         # Aggregate access with the same name and batch add with set union as an optimization.
         # In worst case, all accesses (m) and assignments (n) refer to the same name,
         # the time complexity is O(m x n), this optimizes it as O(m + n).
         scope_name_accesses = defaultdict(set)
-        for (access, enclosing_attribute) in self.__deferred_accesses:
+        for def_access in self.__deferred_accesses:
+            access, enclosing_attribute, enclosing_string_annotation = (
+                def_access.access,
+                def_access.enclosing_attribute,
+                def_access.enclosing_string_annotation,
+            )
             name = ensure_type(access.node, cst.Name).value
             if enclosing_attribute is not None:
                 # if _gen_dotted_names doesn't generate any values, fall back to
@@ -1008,6 +1140,9 @@ class ScopeVisitor(cst.CSTVisitor):
                         name = attr_name
                         break
 
+            if enclosing_string_annotation is not None:
+                access.node = enclosing_string_annotation
+
             scope_name_accesses[(access.scope, name)].add(access)
             access.record_assignments(name)
             access.scope.record_access(name, access)
@@ -1043,70 +1178,3 @@ class ScopeProvider(BatchableMetadataProvider[Optional[Scope]]):
         visitor = ScopeVisitor(self)
         node.visit(visitor)
         visitor.infer_accesses()
-
-
-class QualifiedNameVisitor(cst.CSTVisitor):
-    def __init__(self, provider: "QualifiedNameProvider") -> None:
-        self.provider: QualifiedNameProvider = provider
-
-    def on_visit(self, node: cst.CSTNode) -> bool:
-        scope = self.provider.get_metadata(ScopeProvider, node, None)
-        if scope:
-            self.provider.set_metadata(node, scope.get_qualified_names_for(node))
-        else:
-            self.provider.set_metadata(node, set())
-        super().on_visit(node)
-        return True
-
-
-class QualifiedNameProvider(BatchableMetadataProvider[Collection[QualifiedName]]):
-    """
-    Compute possible qualified names of a variable CSTNode
-    (extends `PEP-3155 <https://www.python.org/dev/peps/pep-3155/>`_).
-    It uses the
-    :func:`~libcst.metadata.Scope.get_qualified_names_for` underlying to get qualified names.
-    Multiple qualified names may be returned, such as when we have conditional imports or an
-    import shadows another. E.g., the provider finds ``a.b``, ``d.e`` and
-    ``f.g`` as possible qualified names of ``c``::
-
-        >>> wrapper = MetadataWrapper(
-        >>>     cst.parse_module(dedent(
-        >>>     '''
-        >>>         if something:
-        >>>             from a import b as c
-        >>>         elif otherthing:
-        >>>             from d import e as c
-        >>>         else:
-        >>>             from f import g as c
-        >>>         c()
-        >>>     '''
-        >>>     ))
-        >>> )
-        >>> call = wrapper.module.body[1].body[0].value
-        >>> wrapper.resolve(QualifiedNameProvider)[call],
-        {
-            QualifiedName(name="a.b", source=QualifiedNameSource.IMPORT),
-            QualifiedName(name="d.e", source=QualifiedNameSource.IMPORT),
-            QualifiedName(name="f.g", source=QualifiedNameSource.IMPORT),
-        }
-
-    For qualified name of a variable in a function or a comprehension, please refer
-    :func:`~libcst.metadata.Scope.get_qualified_names_for` for more detail.
-    """
-
-    METADATA_DEPENDENCIES = (ScopeProvider,)
-
-    def visit_Module(self, node: cst.Module) -> Optional[bool]:
-        visitor = QualifiedNameVisitor(self)
-        node.visit(visitor)
-
-    @staticmethod
-    def has_name(
-        visitor: MetadataDependent, node: cst.CSTNode, name: Union[str, QualifiedName]
-    ) -> bool:
-        """Check if any of qualified name has the str name or :class:`~libcst.metadata.QualifiedName` name."""
-        qualified_names = visitor.get_metadata(QualifiedNameProvider, node, set())
-        if isinstance(name, str):
-            return any(qn.name == name for qn in qualified_names)
-        else:
-            return any(qn == name for qn in qualified_names)
